from __future__ import annotations

import asyncio
import socket
from abc import ABC, abstractmethod
from typing import Generic, Optional, TYPE_CHECKING, TypeVar

import asyncio_dgram

from mcproto.protocol.base_io import BaseAsyncReader, BaseAsyncWriter, BaseSyncReader, BaseSyncWriter

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, Self

    P = ParamSpec("P")

R = TypeVar("R")
T_SOCK = TypeVar("T_SOCK", bound=socket.socket)
T_STREAMREADER = TypeVar("T_STREAMREADER", bound=asyncio.StreamReader)
T_STREAMWRITER = TypeVar("T_STREAMWRITER", bound=asyncio.StreamWriter)
T_DATAGRAM_CLIENT = TypeVar("T_DATAGRAM_CLIENT", bound=asyncio_dgram.aio.DatagramClient)

__all__ = [
    "AsyncConnection",
    "SyncConnection",
    "TCPAsyncConnection",
    "TCPSyncConnection",
    "UDPAsyncConnection",
    "UDPSyncConnection",
]


class SyncConnection(BaseSyncReader, BaseSyncWriter, ABC):
    def __init__(self):
        self.closed = False

    @classmethod
    @abstractmethod
    def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        raise NotImplementedError

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        self._close()
        self.closed = True

    def __enter__(self) -> Self:
        if self.closed:
            raise IOError("Connection already closed.")
        return self

    def __exit__(self, *a, **kw) -> None:
        self.close()


class AsyncConnection(BaseAsyncReader, BaseAsyncWriter, ABC):
    def __init__(self):
        self.closed = False

    @classmethod
    @abstractmethod
    async def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        raise NotImplementedError

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        await self._close()
        self.closed = True

    async def __aenter__(self) -> Self:
        if self.closed:
            raise IOError("Connection already closed.")
        return self

    async def __aexit__(self, *a, **kw) -> None:
        await self.close()


class TCPSyncConnection(SyncConnection, Generic[T_SOCK]):
    def __init__(self, socket: T_SOCK):
        super().__init__()
        self.socket = socket

    @classmethod
    def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        sock = socket.create_connection(address, timeout=timeout)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        return cls(sock)

    def read(self, length: int) -> bytearray:
        """Receive data sent through the connection.

        `length` controls how many bytes we want to receive. If the requested amount
        of bytes isn't available to be received, IOError will be raised.
        """
        result = bytearray()
        while len(result) < length:
            new = self.socket.recv(length - len(result))
            if len(new) == 0:
                # No information at all
                if len(result) == 0:
                    raise IOError("Server did not respond with any information.")
                # Only sent a few bytes, but we requested more
                raise IOError(
                    f"Server stopped responding (got {len(result)} bytes, but expected {length} bytes)."
                    f" Partial obtained data: {result!r}"
                )
            result.extend(new)

        return result

    def write(self, data: bytes) -> None:
        """Send given data over the connection."""
        self.socket.send(data)

    def _close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        self.socket.close()


class TCPAsyncConnection(AsyncConnection, Generic[T_STREAMREADER, T_STREAMWRITER]):
    def __init__(self, reader: T_STREAMREADER, writer: T_STREAMWRITER, timeout: float):
        super().__init__()
        self.reader = reader
        self.writer = writer
        self.timeout = timeout

    @classmethod
    async def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        conn = asyncio.open_connection(address[0], address[1])
        reader, writer = await asyncio.wait_for(conn, timeout=timeout)
        return cls(reader, writer, timeout)

    async def read(self, length: int) -> bytearray:
        """Receive data sent through the connection.

        `length` controls how many bytes we want to receive. If the requested amount
        of bytes isn't available to be received, IOError will be raised.
        """
        result = bytearray()
        while len(result) < length:
            new = await asyncio.wait_for(self.reader.read(length - len(result)), timeout=self.timeout)
            if len(new) == 0:
                # No information at all
                if len(result) == 0:
                    raise IOError("Server did not respond with any information.")
                # Only sent a few bytes, but we requested more
                raise IOError(
                    f"Server stopped responding (got {len(result)} bytes, but expected {length} bytes)."
                    f" Partial obtained data: {result!r}"
                )
            result.extend(new)

        return result

    async def write(self, data: bytes) -> None:
        """Send given data over the connection."""
        self.writer.write(data)

    async def _close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        self.writer.close()

    @property
    def socket(self) -> socket.socket:
        """Obtain the underlying socket behind the asyncio transport."""
        return self.writer.transport._sock  # type: ignore


class UDPSyncConnection(SyncConnection, Generic[T_SOCK]):
    BUFFER_SIZE = 65535

    def __init__(self, socket: T_SOCK, address: tuple[str, int]):
        super().__init__()
        self.socket = socket
        self.address = address

    @classmethod
    def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(timeout)
        return cls(sock, address)

    def read(self, length: Optional[int] = None) -> bytearray:
        """Receive data sent through the connection.

        For UDP connections, `length` parameter is ignored and not required.
        """
        result = bytearray()
        while len(result) == 0:
            received_data, server_addr = self.socket.recvfrom(self.BUFFER_SIZE)
            result.extend(received_data)
        return result

    def write(self, data: bytes) -> None:
        """Send given data over the connection."""
        self.socket.sendto(data, self.address)

    def _close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        self.socket.close()


class UDPAsyncConnection(AsyncConnection, Generic[T_DATAGRAM_CLIENT]):
    def __init__(self, stream: T_DATAGRAM_CLIENT, timeout: float):
        super().__init__()
        self.stream = stream
        self.timeout = timeout

    @classmethod
    async def make_client(cls, address: tuple[str, int], timeout: float) -> Self:
        """Construct a client connection to given address."""
        conn = asyncio_dgram.connect(address)
        stream = await asyncio.wait_for(conn, timeout=timeout)
        return cls(stream, timeout)

    async def read(self, length: Optional[int] = None) -> bytearray:
        """Receive data sent through the connection.

        For UDP connections, `length` parameter is ignored and not required.
        """
        result = bytearray()
        while len(result) == 0:
            received_data, server_addr = await asyncio.wait_for(self.stream.recv(), timeout=self.timeout)
            result.extend(received_data)
        return result

    async def write(self, data: bytes) -> None:
        """Send given data over the connection."""
        await self.stream.send(data)

    async def _close(self) -> None:
        """Close the connection (it cannot be used after this)."""
        self.stream.close()



from __future__ import annotations

import gzip
from collections.abc import Mapping
from typing import TypeVar

from mcproto.buffer import Buffer
from mcproto.packets.abc import Packet
from mcproto.packets.map import PacketMap
from mcproto.protocol.base_io import BaseAsyncReader, BaseAsyncWriter, BaseSyncReader, BaseSyncWriter

T_Packet = TypeVar("T_Packet", bound=Packet)

__all__ = [
    "async_read_packet",
    "async_write_packet",
    "sync_read_packet",
    "sync_write_packet",
]

# PACKET FORMAT:
# | Field name  | Field type    | Notes                                 |
# |-------------|---------------|---------------------------------------|
# | Length      | 32-bit varint | Length (in bytes) of PacketID + Data  |
# | Packet ID   | 32-bit varint |                                       |
# | Data        | byte array    | Internal data to packet of given id   |


# Since the read functions here require PACKET_MAP, we can't move these functions
# directly into BaseWriter/BaseReader classes, as that would be a circular import

PACKET_MAP = PacketMap()


def _serialize_packet(packet: Packet, *, compressed: bool = False) -> Buffer:
    """Serialize the internal packet data, along with it's packet id."""
    packet_data = packet.serialize()

    # Base packet buffer should only contain packet id and internal packet data
    packet_buf = Buffer()
    packet_buf.write_varint(packet.PACKET_ID)
    packet_buf.write(packet_data)

    # If we're serializing a packet as compressed, we compress the packet buffer data
    # and prepend a varint with the size of uncompressed pacekt buffer
    if compressed:
        data_length = len(packet_buf)
        packet_buf = Buffer(gzip.compress(packet_buf))

        data_buf = Buffer()
        data_buf.write_varint(data_length)
        data_buf.write(packet_buf)
        return data_buf
    else:
        return packet_buf


def _deserialize_packet(
    buf: Buffer, *, packet_map: Mapping[int, type[T_Packet]], compressed: bool = False
) -> T_Packet:
    """Deserialize the packet id and it's internal data."""
    if compressed:
        buf.read_varint()  # We don't need this uncompressed length
        compressd_packet_data = buf.read(buf.remaining)
        buf = Buffer(gzip.decompress(compressd_packet_data))

    packet_id = buf.read_varint()
    packet_data = buf.read(buf.remaining)

    return packet_map[packet_id].deserialize(Buffer(packet_data))


def sync_write_packet(writer: BaseSyncWriter, packet: Packet, *, compressed: bool = False) -> None:
    """Write given packet."""
    data_buf = _serialize_packet(packet, compressed=compressed)
    writer.write_bytearray(data_buf)


async def async_write_packet(writer: BaseAsyncWriter, packet: Packet, *, compressed: bool = False) -> None:
    """Write given packet."""
    data_buf = _serialize_packet(packet, compressed=compressed)
    await writer.write_bytearray(data_buf)


def sync_read_packet(
    reader: BaseSyncReader,
    *,
    packet_map: Mapping[int, type[T_Packet]],
    compressed: bool = False,
) -> T_Packet:
    """Read a packet."""
    data_buf = Buffer(reader.read_bytearray())
    return _deserialize_packet(data_buf, packet_map=packet_map, compressed=compressed)


async def async_read_packet(
    reader: BaseAsyncReader,
    *,
    packet_map: Mapping[int, type[T_Packet]],
    compressed: bool = False,
) -> T_Packet:
    """Read a packet."""
    data_buf = Buffer(await reader.read_bytearray())
    return _deserialize_packet(data_buf, packet_map=packet_map, compressed=compressed)



from __future__ import annotations

import inspect
import warnings
from collections.abc import Callable, Iterable
from functools import wraps
from typing import Any, Optional, TYPE_CHECKING, TypeVar, Union, cast, overload

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, Protocol

    P = ParamSpec("P")
    P2 = ParamSpec("P2")
else:
    Protocol = object

T = TypeVar("T")
R = TypeVar("R")
R2 = TypeVar("R2")

__all__ = ["deprecated"]


class DeprecatedReturn(Protocol):
    @overload
    def __call__(self, __x: type[T]) -> type[T]:
        ...

    @overload
    def __call__(self, __x: Callable[P, R]) -> Callable[P, R]:
        ...


@overload
def deprecated(
    obj: Callable[P, R],
    *,
    display_name: Optional[str] = None,
    replacement: Optional[str] = None,
    version: Optional[str] = None,
    date: Optional[str] = None,
    msg: Optional[str] = None,
) -> Callable[P, R]:
    ...


@overload
def deprecated(
    obj: type[T],
    *,
    display_name: Optional[str] = None,
    replacement: Optional[str] = None,
    version: Optional[str] = None,
    date: Optional[str] = None,
    msg: Optional[str] = None,
    methods: Iterable[str],
) -> type[T]:
    ...


@overload
def deprecated(
    obj: None = None,
    *,
    display_name: Optional[str] = None,
    replacement: Optional[str] = None,
    version: Optional[str] = None,
    date: Optional[str] = None,
    msg: Optional[str] = None,
    methods: Optional[Iterable[str]] = None,
) -> DeprecatedReturn:
    ...


def deprecated(
    obj: Any = None,  # noqa: ANN401
    *,
    display_name: Optional[str] = None,
    replacement: Optional[str] = None,
    version: Optional[str] = None,
    date: Optional[str] = None,
    msg: Optional[str] = None,
    methods: Optional[Iterable[str]] = None,
) -> Callable:
    if date is not None and version is not None:
        raise ValueError("Expected removal timeframe can either be a date, or a version, not both.")

    def decorate_func(func: Callable[P2, R2], warn_message: str) -> Callable[P2, R2]:
        @wraps(func)
        def wrapper(*args: P2.args, **kwargs: P2.kwargs) -> R2:
            warnings.warn(warn_message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    @overload
    def decorate(obj: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    def decorate(obj: type[T]) -> type[T]:
        ...

    def decorate(obj: Union[Callable[P, R], type[T]]) -> Union[Callable[P, R], type[T]]:
        # Construct and send the warning message
        name = getattr(obj, "__qualname__", obj.__name__) if display_name is None else display_name
        warn_message = f"'{name}' is deprecated and is expected to be removed"
        if version is not None:
            warn_message += f" in {version}"
        elif date is not None:
            warn_message += f" on {date}"
        else:
            warn_message += " eventually"
        if replacement is not None:
            warn_message += f", use '{replacement}' instead"
        warn_message += "."
        if msg is not None:
            warn_message += f" ({msg})"

        # If we're deprecating class, deprecate it's methods and return the class
        if inspect.isclass(obj):
            obj = cast(type[T], obj)

            if methods is None:
                raise ValueError("When deprecating a class, you need to specify 'methods' which will get the notice")

            for method in methods:
                new_func = decorate_func(getattr(obj, method), warn_message)
                setattr(obj, method, new_func)

            return obj

        # Regular function deprecation
        obj = cast(Callable[P, R], obj)
        if methods is not None:
            raise ValueError("Methods can only be specified when decorating a class, not a function")
        return decorate_func(obj, warn_message)

    # In case the decorator was used like @deprecated, instead of @deprecated()
    # we got the object already, pass it over to the local decorate function
    # This can happen since all of the arguments are optional and can be omitted
    if obj:
        return decorate(obj)
    return decorate



from __future__ import annotations

from enum import IntEnum
from typing import ClassVar, TYPE_CHECKING, Union

from mcproto.buffer import Buffer
from mcproto.packets.abc import GameState, ServerBoundPacket
from mcproto.protocol.base_io import StructFormat

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = [
    "NextState",
    "Handshake",
]


class NextState(IntEnum):
    STATUS = 1
    LOGIN = 2


class Handshake(ServerBoundPacket):
    """Initializes connection between server and client. (Client -> Server)

    :param int protocol_version: Protocol version number to be used.
    :param str server_address: The host/address the client is connecting to.
    :param int server_port: The port the client is connecting to.
    :param Union[NextState, int] next_state: The next state for the server to move into.
    """

    PACKET_ID: ClassVar[int] = 0x00
    GAME_STATE: ClassVar[GameState] = GameState.HANDSHAKING

    __slots__ = ("protocol_version", "server_address", "server_port", "next_state")

    def __init__(
        self,
        *,
        protocol_version: int,
        server_address: str,
        server_port: int,
        next_state: Union[NextState, int],
    ):
        if isinstance(next_state, int):
            rev_lookup = {x.value: x for x in NextState.__members__.values()}
            try:
                next_state = rev_lookup[next_state]
            except KeyError as exc:
                raise ValueError("No such next_state.") from exc

        self.protocol_version = protocol_version
        self.server_address = server_address
        self.server_port = server_port
        self.next_state = next_state

    def serialize(self) -> Buffer:
        buf = Buffer()
        buf.write_varint(self.protocol_version)
        buf.write_utf(self.server_address)
        buf.write_value(StructFormat.USHORT, self.server_port)
        buf.write_varint(self.next_state.value)
        return buf

    @classmethod
    def deserialize(cls, buf: Buffer, /) -> Self:
        return cls(
            protocol_version=buf.read_varint(),
            server_address=buf.read_utf(),
            server_port=buf.read_value(StructFormat.USHORT),
            next_state=buf.read_varint(),
        )



from mcproto.protocol.base_io import BaseAsyncReader, BaseAsyncWriter, BaseSyncReader, BaseSyncWriter, StructFormat

__all__ = [
    "BaseAsyncReader",
    "BaseAsyncWriter",
    "BaseSyncReader",
    "BaseSyncWriter",
    "StructFormat",
]



from __future__ import annotations

import struct
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from enum import Enum
from itertools import count
from typing import Literal, Optional, TYPE_CHECKING, TypeVar, Union, overload

from mcproto.protocol.utils import from_twos_complement, to_twos_complement

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

T = TypeVar("T")
R = TypeVar("R")

__all__ = [
    "BaseAsyncReader",
    "BaseAsyncWriter",
    "BaseSyncReader",
    "BaseSyncWriter",
    "StructFormat",
    "INT_FORMATS_TYPE",
    "FLOAT_FORMATS_TYPE",
]


# region: Format types


class StructFormat(str, Enum):
    """All possible write/read struct types."""

    BOOL = "?"
    CHAR = "c"
    BYTE = "b"
    UBYTE = "B"
    SHORT = "h"
    USHORT = "H"
    INT = "i"
    UINT = "I"
    LONG = "l"
    ULONG = "L"
    FLOAT = "f"
    DOUBLE = "d"
    HALFFLOAT = "e"
    LONGLONG = "q"
    ULONGLONG = "Q"


INT_FORMATS_TYPE: TypeAlias = Union[
    Literal[StructFormat.BYTE],
    Literal[StructFormat.UBYTE],
    Literal[StructFormat.SHORT],
    Literal[StructFormat.USHORT],
    Literal[StructFormat.INT],
    Literal[StructFormat.UINT],
    Literal[StructFormat.LONG],
    Literal[StructFormat.ULONG],
    Literal[StructFormat.LONGLONG],
    Literal[StructFormat.ULONGLONG],
]

FLOAT_FORMATS_TYPE: TypeAlias = Union[
    Literal[StructFormat.FLOAT],
    Literal[StructFormat.DOUBLE],
    Literal[StructFormat.HALFFLOAT],
]

# endregion

# region: Writer classes


class BaseAsyncWriter(ABC):
    """Base class holding asynchronous write buffer/connection interactions."""

    __slots__ = ()

    @abstractmethod
    async def write(self, data: bytes, /) -> None:
        ...

    @overload
    async def write_value(self, fmt: INT_FORMATS_TYPE, value: int, /) -> None:
        ...

    @overload
    async def write_value(self, fmt: FLOAT_FORMATS_TYPE, value: float, /) -> None:
        ...

    @overload
    async def write_value(self, fmt: Literal[StructFormat.BOOL], value: bool, /) -> None:
        ...

    @overload
    async def write_value(self, fmt: Literal[StructFormat.CHAR], value: str, /) -> None:
        ...

    async def write_value(self, fmt: StructFormat, value: object, /) -> None:
        """Write a value of given struct format in big-endian mode."""
        await self.write(struct.pack(">" + fmt.value, value))

    async def _write_varuint(self, value: int, /, *, max_bits: Optional[int] = None) -> None:
        """Write an arbitrarily big unsigned integer in a variable length format.

        This is a standard way of transmitting ints, and it allows smaller numbers to take less bytes.

        Writing will be limited up to integer values of `max_bits` bits, and trying to write bigger values will rase a
        ValueError. Note that setting `max_bits` to for example 32 bits doesn't mean that at most 4 bytes will be sent,
        in this case it would actually take at most 5 bytes, due to the variable encoding overhead.

        Varints send bytes where 7 least significant bits are value bits, and the most significant bit is continuation
        flag bit. If this continuation bit is set (1), it indicates that there will be another varint byte sent after
        this one. The least significant group is written first, followed by each of the more significant groups, making
        varints little-endian, however in groups of 7 bits, not 8.
        """
        value_max = (1 << (max_bits)) - 1 if max_bits is not None else float("inf")
        if value < 0 or value > value_max:
            raise ValueError(f"Tried to write varint outside of the range of {max_bits}-bit int.")

        remaining = value
        while True:
            if remaining & ~0x7F == 0:  # final byte
                await self.write_value(StructFormat.UBYTE, remaining)
                return
            # Write only 7 least significant bits with the first bit being 1, marking there will be another byte
            await self.write_value(StructFormat.UBYTE, remaining & 0x7F | 0x80)
            # Subtract the value we've already sent (7 least significant bits)
            remaining >>= 7

    async def write_varint(self, value: int, /) -> None:
        """Write a 32-bit signed integer in a variable length format.

        For more information about variable length format check `_write_varuint` docstring.
        """
        val = to_twos_complement(value, bits=32)
        await self._write_varuint(val, max_bits=32)

    async def write_varlong(self, value: int, /) -> None:
        """Write a 64-bit signed integer in a variable length format.

        For more information about variable length format check `_write_varuint` docstring.
        """
        val = to_twos_complement(value, bits=64)
        await self._write_varuint(val, max_bits=64)

    async def write_bytearray(self, data: bytes, /) -> None:
        """Write an arbitrary sequence of bytes, prefixed with a varint of it's size."""
        await self.write_varint(len(data))
        await self.write(data)

    async def write_ascii(self, value: str, /) -> None:
        """Write ISO-8859-1 encoded string, with NULL (0x00) at the end to indicate string end."""
        data = bytearray(value, "ISO-8859-1")
        await self.write(data)
        await self.write(bytearray.fromhex("00"))

    async def write_utf(self, value: str, /) -> None:
        """Write a UTF-8 encoded string, prefixed with a varint of it's size (in bytes).

        The maximum amount of UTF-8 characters is limited to 32767.

        Individual UTF-8 characters can take up to 4 bytes, however most of the common ones take up less. Assuming the
        worst case of 4 bytes per every character, at most 131068 data bytes will be written + 3 additional bytes from
        the varint encoding overhead. Note that the limit of 32767 characters holds, even if we're writing shorter
        characters, and we wouldn't cross the 131068 bytes data limit.

        If the given string is longer than this, ValueError will be raised.
        """
        if len(value) > 32767:
            raise ValueError("Maximum character limit for writing strings is 32767 characters.")

        data = bytearray(value, "utf-8")
        await self.write_varint(len(data))
        await self.write(data)

    async def write_optional(self, value: Optional[T], /, writer: Callable[[T], Awaitable[R]]) -> Optional[R]:
        """Writes bool determining is value is present, if it is, also writes the value with writer function.

        When the `value` is None, a bool of False will be written and function will end. Otherwise, if `value` isn't
        None, True will be written, followed by calling the `writer` function wchich will be passed the `value` as the
        only argument. This function is expected to properly write the value into this buffer/connection.

        Will return None if the `value` was None, or the value returned by the `writer` function.
        """
        if value is None:
            await self.write_value(StructFormat.BOOL, False)
            return None

        await self.write_value(StructFormat.BOOL, True)
        return await writer(value)


class BaseSyncWriter(ABC):
    """Base class holding synchronous write buffer/connection interactions."""

    __slots__ = ()

    @abstractmethod
    def write(self, data: bytes, /) -> None:
        ...

    @overload
    def write_value(self, fmt: INT_FORMATS_TYPE, value: int, /) -> None:
        ...

    @overload
    def write_value(self, fmt: FLOAT_FORMATS_TYPE, value: float, /) -> None:
        ...

    @overload
    def write_value(self, fmt: Literal[StructFormat.BOOL], value: bool, /) -> None:
        ...

    @overload
    def write_value(self, fmt: Literal[StructFormat.CHAR], value: str, /) -> None:
        ...

    def write_value(self, fmt: StructFormat, value: object, /) -> None:
        """Write a value of given struct format in big-endian mode."""
        self.write(struct.pack(">" + fmt.value, value))

    def _write_varuint(self, value: int, /, *, max_bits: Optional[int] = None) -> None:
        """Write an arbitrarily big unsigned integer in a variable length format.

        This is a standard way of transmitting ints, and it allows smaller numbers to take less bytes.

        Writing will be limited up to integer values of `max_bits` bits, and trying to write bigger values will rase a
        ValueError. Note that setting `max_bits` to for example 32 bits doesn't mean that at most 4 bytes will be sent,
        in this case it would actually take at most 5 bytes, due to the variable encoding overhead.

        Varints send bytes where 7 least significant bits are value bits, and the most significant bit is continuation
        flag bit. If this continuation bit is set (1), it indicates that there will be another varint byte sent after
        this one. The least significant group is written first, followed by each of the more significant groups, making
        varints little-endian, however in groups of 7 bits, not 8.
        """
        value_max = (1 << (max_bits)) - 1 if max_bits is not None else float("inf")
        if value < 0 or value > value_max:
            raise ValueError(f"Tried to write varint outside of the range of {max_bits}-bit int.")

        remaining = value
        while True:
            if remaining & ~0x7F == 0:  # final byte
                self.write_value(StructFormat.UBYTE, remaining)
                return
            # Write only 7 least significant bits with the first bit being 1, marking there will be another byte
            self.write_value(StructFormat.UBYTE, remaining & 0x7F | 0x80)
            # Subtract the value we've already sent (7 least significant bits)
            remaining >>= 7

    def write_varint(self, value: int, /) -> None:
        """Write a 32-bit signed integer in a variable length format.

        For more information about variable length format check `_write_varuint` docstring.
        """
        val = to_twos_complement(value, bits=32)
        self._write_varuint(val, max_bits=32)

    def write_varlong(self, value: int, /) -> None:
        """Write a 64-bit signed integer in a variable length format.

        For more information about variable length format check `_write_varuint` docstring.
        """
        val = to_twos_complement(value, bits=64)
        self._write_varuint(val, max_bits=64)

    def write_bytearray(self, data: bytes, /) -> None:
        """Write an arbitrary sequence of bytes, prefixed with a varint of it's size."""
        self.write_varint(len(data))
        self.write(data)

    def write_ascii(self, value: str, /) -> None:
        """Write ISO-8859-1 encoded string, with NULL (0x00) at the end to indicate string end."""
        data = bytearray(value, "ISO-8859-1")
        self.write(data)
        self.write(bytearray.fromhex("00"))

    def write_utf(self, value: str, /) -> None:
        """Write a UTF-8 encoded string, prefixed with a varint of it's size (in bytes).

        The maximum amount of UTF-8 characters is limited to 32767.

        Individual UTF-8 characters can take up to 4 bytes, however most of the common ones take up less. Assuming the
        worst case of 4 bytes per every character, at most 131068 data bytes will be written + 3 additional bytes from
        the varint encoding overhead. Note that the limit of 32767 characters holds, even if we're writing shorter
        characters, and we wouldn't cross the 131068 bytes data limit.

        If the given string is longer than this, ValueError will be raised.
        """
        if len(value) > 32767:
            raise ValueError("Maximum character limit for writing strings is 32767 characters.")

        data = bytearray(value, "utf-8")
        self.write_varint(len(data))
        self.write(data)

    def write_optional(self, value: Optional[T], /, writer: Callable[[T], R]) -> Optional[R]:
        """Writes bool determining is value is present, if it is, also writes the value with writer function.

        When the `value` is None, a bool of False will be written and function will end. Otherwise, if `value` isn't
        None, True will be written, followed by calling the `writer` function wchich will be passed the `value` as the
        only argument. This function is expected to properly write the value into this buffer/connection.

        Will return None if the `value` was None, or the value returned by the `writer` function.
        """
        if value is None:
            self.write_value(StructFormat.BOOL, False)
            return None

        self.write_value(StructFormat.BOOL, True)
        return writer(value)


# endregion
# region: Reader classes


class BaseAsyncReader(ABC):
    """Base class holding asynchronous read buffer/connection interactions."""

    __slots__ = ()

    @abstractmethod
    async def read(self, length: int, /) -> bytearray:
        ...

    @overload
    async def read_value(self, fmt: INT_FORMATS_TYPE, /) -> int:
        ...

    @overload
    async def read_value(self, fmt: FLOAT_FORMATS_TYPE, /) -> float:
        ...

    @overload
    async def read_value(self, fmt: Literal[StructFormat.BOOL], /) -> bool:
        ...

    @overload
    async def read_value(self, fmt: Literal[StructFormat.CHAR], /) -> str:
        ...

    async def read_value(self, fmt: StructFormat, /) -> object:
        """Read a value into given struct format in big-endian mode.

        The amount of bytes to read will be determined based on the struct format automatically.
        """
        length = struct.calcsize(fmt.value)
        data = await self.read(length)
        unpacked = struct.unpack(">" + fmt.value, data)
        return unpacked[0]

    async def _read_varuint(self, *, max_bits: Optional[int] = None) -> int:
        """Read an arbitrarily big unsigned integer in a variable length format.

        This is a standard way of transmitting ints, and it allows smaller numbers to take less bytes.

        Reading will be limited up to integer values of `max_bits` bits, and trying to read bigger values will rase an
        IOError. Note that setting `max_bits` to for example 32 bits doesn't mean that at most 4 bytes will be read,
        in this case it would actually read at most 5 bytes, due to the variable encoding overhead.

        Varints send bytes where 7 least significant bits are value bits, and the most significant bit is continuation
        flag bit. If this continuation bit is set (1), it indicates that there will be another varint byte sent after
        this one. The least significant group is written first, followed by each of the more significant groups, making
        varints little-endian, however in groups of 7 bits, not 8.
        """
        value_max = (1 << (max_bits)) - 1 if max_bits is not None else float("inf")

        result = 0
        for i in count():
            byte = await self.read_value(StructFormat.UBYTE)
            # Read 7 least significant value bits in this byte, and shift them appropriately to be in the right place
            # then simply add them (OR) as additional 7 most significant bits in our result
            result |= (byte & 0x7F) << (7 * i)

            # Ensure that we stop reading and raise an error if the size gets over the maximum
            # (if the current amount of bits is higher than allowed size in bits)
            if result > value_max:
                raise IOError(f"Received varint was outside the range of {max_bits}-bit int.")

            # If the most significant bit is 0, we should stop reading
            if not byte & 0x80:
                break

        return result

    async def read_varint(self) -> int:
        """Read a 32-bit signed integer in a variable length format.

        For more information about variable length format check `_read_varuint` docstring.
        """
        unsigned_num = await self._read_varuint(max_bits=32)
        val = from_twos_complement(unsigned_num, bits=32)
        return val

    async def read_varlong(self) -> int:
        """Read a 64-bit signed integer in a variable length format.

        For more information about variable length format check `_read_varuint` docstring.
        """
        unsigned_num = await self._read_varuint(max_bits=64)
        val = from_twos_complement(unsigned_num, bits=64)
        return val

    async def read_bytearray(self, /) -> bytearray:
        """Read an arbitrary sequence of bytes, prefixed with a varint of it's size."""
        length = await self.read_varint()
        return await self.read(length)

    async def read_ascii(self) -> str:
        """Read ISO-8859-1 encoded string, until we encounter NULL (0x00) at the end indicating string end."""
        # Keep reading bytes until we find NULL
        result = bytearray()
        while len(result) == 0 or result[-1] != 0:
            byte = await self.read(1)
            result.extend(byte)
        return result[:-1].decode("ISO-8859-1")

    async def read_utf(self) -> str:
        """Read a UTF-8 encoded string, prefixed with a varint of it's size (in bytes).

        The maximum amount of UTF-8 characters is limited to 32767.

        Individual UTF-8 characters can take up to 4 bytes, however most of the common ones take up less. Assuming the
        worst case of 4 bytes per every character, at most 131068 data bytes will be read + 3 additional bytes from
        the varint encoding overhead. Note that the while the limit of 32767 characters is enforced, this check only
        happens after the data was already read.

        If the given string is longer than this, IOError will be raised.
        """
        length = await self.read_varint()
        if length > 131068:
            raise IOError(f"Maximum read limit for utf strings is 131068 bytes, got {length}.")

        data = await self.read(length)
        chars = data.decode("utf-8")

        if len(chars) > 32767:
            raise IOError(f"Maximum read limit for utf strings is 32767 characters, got {len(chars)}.")

        return chars

    async def read_optional(self, reader: Callable[[], Awaitable[R]]) -> Optional[R]:
        """Reads bool determining is value is present, if it is, also reads the value with reader function.

        When False is read, the function will not read anything and end. Otherwise, if True is read, the `reader`
        function will be called, which is expected to properly read the value from this buffer/connection.

        Will return None if the False was encountered, or the value returned by the `reader` function.
        """
        if not await self.read_value(StructFormat.BOOL):
            return None

        return await reader()


class BaseSyncReader(ABC):
    """Base class holding synchronous read buffer/connection interactions."""

    __slots__ = ()

    @abstractmethod
    def read(self, length: int, /) -> bytearray:
        ...

    @overload
    def read_value(self, fmt: INT_FORMATS_TYPE, /) -> int:
        ...

    @overload
    def read_value(self, fmt: FLOAT_FORMATS_TYPE, /) -> float:
        ...

    @overload
    def read_value(self, fmt: Literal[StructFormat.BOOL], /) -> bool:
        ...

    @overload
    def read_value(self, fmt: Literal[StructFormat.CHAR], /) -> str:
        ...

    def read_value(self, fmt: StructFormat, /) -> object:
        """Read a value into given struct format in big-endian mode.

        The amount of bytes to read will be determined based on the struct format automatically.
        """
        length = struct.calcsize(fmt.value)
        data = self.read(length)
        unpacked = struct.unpack(">" + fmt.value, data)
        return unpacked[0]

    def _read_varuint(self, *, max_bits: Optional[int] = None) -> int:
        """Read an arbitrarily big unsigned integer in a variable length format.

        This is a standard way of transmitting ints, and it allows smaller numbers to take less bytes.

        Reading will be limited up to integer values of `max_bits` bits, and trying to read bigger values will rase an
        IOError. Note that setting `max_bits` to for example 32 bits doesn't mean that at most 4 bytes will be read,
        in this case it would actually read at most 5 bytes, due to the variable encoding overhead.

        Varints send bytes where 7 least significant bits are value bits, and the most significant bit is continuation
        flag bit. If this continuation bit is set (1), it indicates that there will be another varint byte sent after
        this one. The least significant group is written first, followed by each of the more significant groups, making
        varints little-endian, however in groups of 7 bits, not 8.
        """
        value_max = (1 << (max_bits)) - 1 if max_bits is not None else float("inf")

        result = 0
        for i in count():
            byte = self.read_value(StructFormat.UBYTE)
            # Read 7 least significant value bits in this byte, and shift them appropriately to be in the right place
            # then simply add them (OR) as additional 7 most significant bits in our result
            result |= (byte & 0x7F) << (7 * i)

            # Ensure that we stop reading and raise an error if the size gets over the maximum
            # (if the current amount of bits is higher than allowed size in bits)
            if result > value_max:
                raise IOError(f"Received varint was outside the range of {max_bits}-bit int.")

            # If the most significant bit is 0, we should stop reading
            if not byte & 0x80:
                break

        return result

    def read_varint(self) -> int:
        """Read a 32-bit signed integer in a variable length format.

        For more information about variable length format check `_read_varuint` docstring.
        """
        unsigned_num = self._read_varuint(max_bits=32)
        val = from_twos_complement(unsigned_num, bits=32)
        return val

    def read_varlong(self) -> int:
        """Read a 64-bit signed integer in a variable length format.

        For more information about variable length format check `_read_varuint` docstring.
        """
        unsigned_num = self._read_varuint(max_bits=64)
        val = from_twos_complement(unsigned_num, bits=64)
        return val

    def read_bytearray(self) -> bytearray:
        """Read an arbitrary sequence of bytes, prefixed with a varint of it's size."""
        length = self.read_varint()
        return self.read(length)

    def read_ascii(self) -> str:
        """Read ISO-8859-1 encoded string, until we encounter NULL (0x00) at the end indicating string end."""
        # Keep reading bytes until we find NULL
        result = bytearray()
        while len(result) == 0 or result[-1] != 0:
            byte = self.read(1)
            result.extend(byte)
        return result[:-1].decode("ISO-8859-1")

    def read_utf(self) -> str:
        """Read a UTF-8 encoded string, prefixed with a varint of it's size (in bytes).

        The maximum amount of UTF-8 characters is limited to 32767.

        Individual UTF-8 characters can take up to 4 bytes, however most of the common ones take up less. Assuming the
        worst case of 4 bytes per every character, at most 131068 data bytes will be read + 3 additional bytes from
        the varint encoding overhead. Note that the while the limit of 32767 characters is enforced, this check only
        happens after the data was already read.

        If the given string is longer than this, IOError will be raised.
        """
        length = self.read_varint()
        if length > 131068:
            raise IOError(f"Maximum read limit for utf strings is 131068 bytes, got {length}.")

        data = self.read(length)
        chars = data.decode("utf-8")

        if len(chars) > 32767:
            raise IOError(f"Maximum read limit for utf strings is 32767 characters, got {len(chars)}.")

        return chars

    def read_optional(self, reader: Callable[[], R]) -> Optional[R]:
        """Reads bool determining is value is present, if it is, also reads the value with reader function.

        When False is read, the function will not read anything and end. Otherwise, if True is read, the `reader`
        function will be called, which is expected to properly read the value from this buffer/connection.

        Will return None if the False was encountered, or the value returned by the `reader` function.
        """
        if not self.read_value(StructFormat.BOOL):
            return None

        return reader()


# endregion



from mcproto.protocol.base_io import BaseSyncReader, BaseSyncWriter

__all__ = ["Buffer"]


class Buffer(BaseSyncWriter, BaseSyncReader, bytearray):
    __slots__ = ("pos",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pos = 0

    def write(self, data: bytes) -> None:
        """Write new data into the buffer."""
        self.extend(data)

    def read(self, length: int) -> bytearray:
        """Read data stored in the buffer.

        Reading data doesn't remove that data, rather that data is treated as already read, and
        next read will start from the first unread byte. If freeing the data is necessary, check the clear function.

        Trying to read more data than is available will raise an IOError, however it will deplete the remaining data
        and the partial data that was read will be a part of the error message. This behavior is here to mimic reading
        from a socket connection.
        """
        end = self.pos + length

        if end > len(self):
            data = self[self.pos : len(self)]
            bytes_read = len(self) - self.pos
            self.pos = len(self)
            raise IOError(
                "Requested to read more data than available."
                f" Read {bytes_read} bytes: {data}, out of {length} requested bytes."
            )

        try:
            return self[self.pos : end]
        finally:
            self.pos = end

    def clear(self, only_already_read: bool = False) -> None:
        """
        Clear out the stored data and reset position.

        If `only_already_read` is True, only clear out the data which was already read, and reset the position.
        This is mostly useful to avoid keeping large chunks of data in memory for no reason.
        """
        if only_already_read:
            del self[: self.pos]
        else:
            super().clear()
        self.pos = 0

    def reset(self) -> None:
        """Reset the position in the buffer."""
        self.pos = 0

    def flush(self) -> bytearray:
        """Read all of the remaining data in the buffer and clear it out."""
        data = self[self.pos : len(self)]
        self.clear()
        return data

    @property
    def remaining(self) -> int:
        """Get the amount of bytes that's still remaining in be buffer to be read."""
        return len(self) - self.pos



from __future__ import annotations

__all__ = ["to_twos_complement", "from_twos_complement"]


def to_twos_complement(num: int, bits: int) -> int:
    """Convert a given number into twos complement format of given amount of bits."""
    value_max = 1 << (bits - 1)
    value_min = value_max * -1
    # With two's complement, we have one more negative number than positive
    # this means we can't be exactly at value_max, but we can be at exactly value_min
    if num >= value_max or num < value_min:
        raise ValueError(f"Can't convert number {num} into {bits}-bit twos complement format - out of range")

    return num + (1 << bits) if num < 0 else num


def from_twos_complement(num: int, bits: int) -> int:
    """Convert a given number from twos complement format of given amount of bits."""
    value_max = (1 << bits) - 1
    if num < 0 or num > value_max:
        raise ValueError(f"Can't convert number {num} from {bits}-bit twos complement format - out of range")

    if num & (1 << (bits - 1)) != 0:
        num -= 1 << bits

    return num



from __future__ import annotations

import importlib
import pkgutil
import warnings
from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterator, Sequence
from types import ModuleType
from typing import Any, ClassVar, Generic, NamedTuple, NoReturn, TYPE_CHECKING, TypeVar

from mcproto.utils.abc import RequiredParamsABCMixin

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

K = TypeVar("K", bound=Hashable)
V = TypeVar("V")

__all__ = ["VersionMap", "WalkableModuleData"]


class WalkableModuleData(NamedTuple):
    module: ModuleType
    info: pkgutil.ModuleInfo
    member_names: Sequence[str]


class VersionMap(ABC, RequiredParamsABCMixin, Generic[K, V]):
    """Base class for map classes that allow obtaining versioned implementations for same components.

    Some components (classes) need to be versioned as they might change between different protocol
    versions. VersionMap classes are responsible for finding the implementation for given component
    that matches the requested version for.

    If the exact requested version is not found, the closest older/lesser version will be used instead,
    however as this might cause issues, a user warning will be produced alongside this happening.

    NOTE: If you need typed implementations, you are advised to instead import the versioned class
    directly, as due to the dynamic nature of this map, and the changing nature of the versioned
    implementations, no type information can be provided here.

    WARNING: You are not expected to initialize these mapping classes! That should be left for the
    library to handle and provide a map for each individual versioned components.
    """

    # TODO: Describe the obtain key, lazy loading and preload_all in init

    # TODO: Look into generating stubs somehow to preserve the type information even with how
    # dynamic this class is. (Could require making this class generic over a literal int, being
    # the version).

    _REQUIRED_CLASS_VARS: ClassVar[Sequence[str]] = [
        "SUPPORTED_VERSIONS",
        "_SEARCH_DIR_QUALNAME",
        "COMPATIBLE_FALLBACK_VERSIONS",
    ]

    SUPPORTED_VERSIONS: ClassVar[set[int]]
    COMPATIBLE_FALLBACK_VERSIONS: ClassVar[set[int]]
    _SEARCH_DIR_QUALNAME: ClassVar[str]

    __slots__ = ("_version_map",)

    def __init__(self, preload_all: bool = False):
        self._version_map: dict[int, dict[K, V]] = {}

        if preload_all:
            for version in self.SUPPORTED_VERSIONS:
                self.load_version(version)

    def load_version(self, protocol_version: int) -> None:
        """Load all components for given protocol version."""
        for key, member in self._walk_valid_components(protocol_version):
            dct = self._version_map.setdefault(protocol_version, {})

            if key in dct:
                raise ValueError(f"Value {member} already registered (key collision: {key})")

            dct[key] = member

    def obtain(self, protocol_version: int, key: K) -> V:
        """Obtain component for given protocol version, that matches the obtain key."""
        protocol_version = self._get_supported_protocol_version(protocol_version)

        # Lazy loading: if this protocol version isn't in the version map, it wasn't yet loaded, do so now.
        if protocol_version not in self._version_map:
            self.load_version(protocol_version)

        return self._version_map[protocol_version][key]

    def make_version_map(self, protocol_version: int) -> dict[K, V]:
        """Construct a dictionary mapping (obtain key -> value) for values of given protocol version."""
        protocol_version = self._get_supported_protocol_version(protocol_version)

        # Lazy loading: if this protocol version isn't in the version map, it wasn't yet loaded, do so now.
        if protocol_version not in self._version_map:
            self.load_version(protocol_version)

        return self._version_map[protocol_version].copy()

    @abstractmethod
    def _check_obj(
        self, obj: Any, module_data: WalkableModuleData, protocol_version: int  # noqa: ANN401
    ) -> TypeGuard[V]:
        """Determine whether a member object should be considered as a valid component for given protocol version.

        This method will be called for each potential member object (found when walking over all members of
        __all__ from all submodules of the package for given protocol version).

        This function shouldn't include any checks on whether an object is already registered in the version map
        (key collisions), these are handled during the collection in load_version, all this function is responsible
        for is checking whether this object is a valid component, components with conflicting keys are still
        considered valid here, as they're handled elsewhere.

        Although if there is some additional data that needs to be unique for a component to be valid, which wouldn't
        be caught as a key collision, this function can raise a ValueError.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _make_obtain_key(cls, obj: V, module_data: WalkableModuleData, protocol_version: int) -> K:
        """Construct a unique obtain key for given object under given protocol version.


        Note: While the protocol version might be beneficial to know when constructing
        the obtain key, it shouldn't be used directly as a part of the key, as the items
        will already be split by their protocol versions, and this version will be known
        at obtaining time.
        """
        raise NotImplementedError

    @classmethod
    def _get_package_module(cls, protocol_version: int) -> ModuleType:
        """Obtain the package module containing all components specific to given protocol version.

        This relies on the _SEARCH_DIR_QUALNAME and naming scheme of v{protocol_version} for the modules.
        """
        module_name = f"{cls._SEARCH_DIR_QUALNAME}.v{protocol_version}"
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as exc:
            raise ValueError(
                f"Protocol version ({protocol_version}) isn't supported, yet it was listed as supported. Report this!"
            ) from exc

        return module

    @classmethod
    def _walk_submodules(cls, module: ModuleType) -> Iterator[WalkableModuleData]:
        """Go over all submodules of given module, that properly specify __all__.

        The yielded modules are expected to contain the versioned component items.

        If a submodule that doesn't define __all__ is present, it will be skipped, as we don't
        consider it walkable. (Walking over all variables defined in this module isn't viable,
        since it would include imports etc. making defining __all__ a requirement.)
        """

        def on_error(name: str) -> NoReturn:
            raise ImportError(name=name)

        for module_info in pkgutil.walk_packages(module.__path__, f"{module.__name__}.", onerror=on_error):
            imported_module = importlib.import_module(module_info.name)

            if not hasattr(imported_module, "__all__"):
                continue
            member_names = imported_module.__all__

            if not isinstance(member_names, Sequence):
                raise TypeError(f"Module {module_info.name!r}'s __all__ isn't defined as a sequence.")

            for member_name in member_names:
                if not isinstance(member_name, str):
                    raise TypeError(f"Module {module_info.name!r}'s __all__ contains non-string item.")

            yield WalkableModuleData(imported_module, module_info, member_names)

    @classmethod
    def _walk_members(cls, module_data: WalkableModuleData) -> Iterator[object]:
        """Go over all members specified in __all__ of given walkable (sub)module.

        Yields all
        """
        for member_name in module_data.member_names:
            try:
                member = getattr(module_data.module, member_name)
            except AttributeError as exc:
                module_name = module_data.info.name
                raise TypeError(f"Member {member_name!r} of {module_name!r} module isn't defined.") from exc

            yield member

    def _walk_valid_components(self, protocol_version: int) -> Iterator[tuple[K, V]]:
        """Go over all components/members belonging to given protocol version."""
        version_module = self._get_package_module(protocol_version)
        for module_data in self._walk_submodules(version_module):
            for member in self._walk_members(module_data):
                if self._check_obj(member, module_data, protocol_version):
                    key = self._make_obtain_key(member, module_data, protocol_version)
                    yield key, member

    def _get_supported_protocol_version(self, protocol_version: int) -> int:
        """Given a protocol version, return closest older supported version, or the version itself.

        - If given protocol version is one of supported versions, this function will simply return it.
        - Otherwise, search for the closest older version to the requested version
            - If there is no older version in the supported version, ValueError is raised
            - If the requested version is in COMPATIBLE_FALLBACK_VERSIONS, this fallback will be
              considered safe, and no errors/warnings will be produced.
            - Otherwise, the closest version will still be returned, but a UserWarning will be emitted
              mentioning that this fallback occurred and that the version might not be fully compatible.
        """

        if protocol_version in self.SUPPORTED_VERSIONS:
            return protocol_version

        # Pick the most recent, but older version to the one that was requested
        older_versions = filter(lambda version: version < protocol_version, self.SUPPORTED_VERSIONS)
        try:
            protocol_version_closest = max(older_versions)
        except ValueError as exc:  # No older protocol version found
            raise ValueError(
                f"Requested protocol version ({protocol_version}) isn't supported"
                ", and no older version to fall back to was found."
            ) from exc

        # Since using an older/lesser version to the one that was requested is a fallback behavior and
        # could easily cause issues, emit a warning here, unless the version was explicitly marked compatible.
        if protocol_version not in self.COMPATIBLE_FALLBACK_VERSIONS:
            warnings.warn(
                f"Falling back to older protocol version {protocol_version_closest}, "
                f"as the requested version ({protocol_version}) isn't supported.",
                category=UserWarning,
            )

        return protocol_version_closest



from __future__ import annotations

from collections.abc import Sequence
from enum import IntEnum
from typing import ClassVar

from mcproto.utils.abc import RequiredParamsABCMixin, Serializable

__all__ = [
    "GameState",
    "PacketDirection",
    "Packet",
    "ServerBoundPacket",
    "ClientBoundPacket",
]


class GameState(IntEnum):
    HANDSHAKING = 0
    STATUS = 1
    LOGIN = 2
    PLAY = 3


class PacketDirection(IntEnum):
    SERVERBOUND = 0
    CLIENTBOUND = 1


class Packet(Serializable, RequiredParamsABCMixin):
    """Base class for all packets"""

    _REQUIRED_CLASS_VARS: ClassVar[Sequence[str]] = ["PACKET_ID", "GAME_STATE"]
    _REQUIRED_CLASS_VARS_NO_MRO: ClassVar[Sequence[str]] = ["__slots__"]

    __slots__ = ()

    PACKET_ID: ClassVar[int]
    GAME_STATE: ClassVar[GameState]


class ServerBoundPacket(Packet):
    """Packet bound to a server (Client -> Server)."""


class ClientBoundPacket(Packet):
    """Packet bound to a client (Server -> Client)."""



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import ClassVar, TYPE_CHECKING

from mcproto.buffer import Buffer

if TYPE_CHECKING:
    from typing_extensions import Self


class RequiredParamsABCMixin:
    """Mixin class to ABCs that require certain attributes to be set in order to allow initialization.

    This class performs a similar check to what ABCs already do for abstractmethods, but for class
    variables. The required class variable names are set with _REQUIRED_CLASS_VARS class variable,
    which itself is automatically required.

    Just like with ABCs, this doesn't prevent creation of classes without these required class vars
    missing, only initialization is prevented. This is done to allow creation of a more specific, but
    still abstract class.

    Additionally, you can also define _REQUIRED_CLASS_VARS_NO_MRO class var, holding names of class vars
    which should be defined on given class directly. That means inheritance will be ignored so even if a
    subclass defines the required class var, unless the latest class also defines it, this check will fail.

    This is often useful for classes that are expected to be slotted, as each subclass need to define
    __slots__, otherwise a __dict__ will automatically be made for it. However this is entirely optional,
    and if _REQUIRED_CLASS_VARS_NO_MRO isn't set, no such check will be performed.
    """

    _REQUIRRED_CLASS_VARS: ClassVar[Sequence[str]]
    _REQUIRED_CLASS_VARS_NO_MRO: ClassVar[Sequence[str]]

    def __new__(cls: type[Self], *a, **kw) -> Self:
        """Enforce required parameters being set for each instance of the concrete classes."""
        _err_msg = f"Can't instantiate abstract {cls.__name__} class without defining " + "{!r} classvar"

        _required_class_vars = getattr(cls, "_REQUIRED_CLASS_VARS", None)
        if _required_class_vars is None:
            raise TypeError(_err_msg.format("_REQUIRED_CLASS_VARS"))

        for req_attr in _required_class_vars:
            if not hasattr(cls, req_attr):
                raise TypeError(_err_msg.format(req_attr))

        _required_class_vars_no_mro = getattr(cls, "_REQUIRED_CLASS_VARS_NO_MRO", None)
        if _required_class_vars_no_mro is None:
            return super().__new__(cls)

        for req_no_mro_attr in _required_class_vars_no_mro:
            if req_no_mro_attr not in vars(cls):
                emsg = _err_msg.format(req_no_mro_attr) + " explicitly"
                if hasattr(cls, req_no_mro_attr):
                    emsg += f"({req_no_mro_attr} found in a subclass, but not explicitly in {cls.__name__})"
                raise TypeError(emsg)

        return super().__new__(cls)


class Serializable(ABC):
    """Base class for any type that should be (de)serializable into/from given buffer data."""

    @abstractmethod
    def serialize(self) -> Buffer:
        """Represent the object as a transmittable sequence of bytes."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def deserialize(cls, buf: Buffer, /) -> Self:
        """Construct the object from it's received byte representation."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mcproto',
 'mcproto.packets',
 'mcproto.packets.v757',
 'mcproto.packets.v757.handshaking',
 'mcproto.packets.v757.status',
 'mcproto.protocol',
 'mcproto.utils']

package_data = \
{'': ['*']}

install_requires = \
['asyncio-dgram>=2.1.2,<3.0.0']

setup_kwargs = {
    'name': 'mcproto',
    'version': '0.1.2',
    'description': 'Library providing easy interactions with minecraft sevrers',
    'long_description': '# <img src="https://i.imgur.com/nPCcxts.png" height=25> McProto\n\n[![discord chat](https://img.shields.io/discord/936788458939224094.svg?logo=Discord)](https://discord.gg/C2wX7zduxC)\n![supported python versions](https://img.shields.io/pypi/pyversions/mcproto.svg)\n[![current PyPI version](https://img.shields.io/pypi/v/mcproto.svg)](https://pypi.org/project/mcproto/)\n[![Validation](https://github.com/ItsDrike/mcproto/actions/workflows/validation.yml/badge.svg)](https://github.com/ItsDrike/mcproto/actions/workflows/validation.yml)\n[![Unit tests](https://github.com/ItsDrike/mcproto/actions/workflows/unit-tests.yml/badge.svg)](https://github.com/ItsDrike/mcproto/actions/workflows/unit-tests.yml)\n\nThis is a heavily Work-In-Progress library, which attempts to be a full wrapper around the minecraft protocol, allowing\nfor simple interactions with minecraft servers, and perhaps even for use as a base to a full minecraft server\nimplementation in python (though the speed will very likely be quite terrible, making it probably unusable as any\nreal-world playable server).\n\nCurrently, the library is very limited and doesn\'t yet have any documentation, so while contributions are welcome, fair\nwarning that there is a lot to comprehend in the code-base and it may be challenging to understand it all.\n\n## Installation\n\nMcproto is available on [PyPI](https://pypi.org/project/mcproto/), and can be installed trivially with:\n\n```bash\npython3 -m pip install mcproto\n```\n\nThat said, as mcproto is still in development, the PyPI version will likely go out of date quite soon. This means that\nit may lack some already implemented features, or contain already fixed bugs. For that reason, you can also consider\ninstalling mcproto through git, to get the most recent version. But know that while this will mean you\'ll be getting\nall of the new features, this version is much more likely to contain bugs than the one on PyPI, so make your decision\nwisely. To install the latest version from git, you can use the command below:\n\n```bash\npython3 -m pip install \'mcproto @ git+https://github.com/py-mine/mcproto@main\'\n```\n\nAlternatively, if you want to poke around with mcproto\'s code, you can always include mcproto as a full [git\nsubmodule](https://git-scm.com/book/en/v2/Git-Tools-Submodules) to your project, or just git clone it directly and play\naround with it from REPL.\n\n## Examples\n\nSince there is no documentation, to satisfy some curious minds that really want to use this library even in this\nheavily unfinished state, here\'s a few simple snippets of it in practice:\n\n### Manual communication with the server\n\nAs sending entire packets is still being worked on, the best solution to communicate with a server is to send the data\nmanually, using our connection reader/writer, and buffers (being readers/writers, but only from bytearrays as opposed\nto using an actual connection).\n\nFair warning: This example is pretty long, but that\'s because it aims to explain the minecraft protocol to people that\nsee it for the first time, and so a lot of explanation comments are included. But the code itself is actually quite\nsimple, due to a bunch of helpful read/write methods the library already provides.\n\n```python\nimport json\nimport asyncio\n\nfrom mcproto.buffer import Buffer\nfrom mcproto.connection import TCPAsyncConnection\nfrom mcproto.protocol.base_io import StructFormat\n\n\nasync def handshake(conn: TCPAsyncConnection, ip: str, port: int = 25565) -> None:\n    # As a simple example, let\'s request status info from a server.\n    # (this is what you see in the multiplayer server list, i.e. the server\'s motd, icon, info\n    # about how many players are connected, etc.)\n\n    # To do this, we first need to understand how are minecraft packets composed, and take a look\n    # at the specific packets that we\'re interested in. Thankfully, there\'s an amazing community\n    # made wiki that documents all of this! You can find it at https://wiki.vg/\n\n    # Alright then, let\'s take a look at the (uncompressed) packet format specification:\n    # https://wiki.vg/Protocol#Packet_format\n    # From the wiki, we can see that a packet is composed of 3 fields:\n    # - Packet length (in bytes), sent as a variable length integer\n    #       combined length of the 2 fields below\n    # - Packet ID, also sent as varint\n    #       each packet has a unique number, that we use to find out which packet it is\n    # - Data, specific to the individual packet\n    #       every packet can hold different kind of data, this will be shown in the packet\'s\n    #       specification (you can find these in wiki.vg)\n\n    # Ok then, with this knowledge, let\'s establish a connection with our server, and request\n    # status. To do this, we fist need to send a handshake packet. Let\'s do it:\n\n    # Let\'s take a look at what data the Handshake packet should contain:\n    # https://wiki.vg/Protocol#Handshake\n    handshake = Buffer()\n    # We use 47 for the protocol version, as it\'s quite old, and will work with almost any server\n    handshake.write_varint(47)\n    handshake.write_utf(ip)\n    handshake.write_value(StructFormat.USHORT, port)\n    handshake.write_varint(1)  # Intention to query status\n\n    # Nice! Now that we have the packet data, let\'s follow the packet format and send it.\n    # Let\'s prepare another buffer that will contain the last 2 fields (packet id and data)\n    # combined. We do this since the first field will require us to know the size of these\n    # two combined, so let\'s just put them into 1 buffer.\n    packet = Buffer()\n    packet.write_varint(0)  # Handshake packet has packet ID of 0\n    packet.write(handshake)  # Full data from our handshake packet\n\n    # And now, it\'s time to send it!\n    await conn.write_varint(len(packet))  # First field (size of packet id + data)\n    await conn.write(packet)  # Second + Third fileds (packet id + data)\n\n\nasync def status(conn: TCPAsyncConnection, ip: str, port: int = 25565) -> dict:\n    # This function will be called right after a handshake\n    # Sending this packet told the server recognize our connection, and since we\'ve specified\n    # the intention to query status, it then moved us to STATUS game state.\n\n    # Different game states have different packets that we can send out, for example there is a\n    # game state for login, that we\'re put into while joining the server, and from it, we tell\n    # the server our username player UUID, etc.\n\n    # The packet IDs are unique to each game state, so since we\'re now in status state, a packet\n    # with ID of 0 is no longer the handshake packet, but rather the status request packet\n    # (precisely what we need).\n    # https://wiki.vg/Protocol#Status_Request\n\n    # The status request packet is empty, and doesn\'t contain any data, it just instructs the\n    # server to send us back a status response packet. Let\'s send it!\n    packet = Buffer()\n    packet.write_varint(0)  # Status request packet ID\n\n    await conn.write_varint(len(packet))\n    await conn.write(packet)\n\n    # Now, let\'s receive the response packet from the server\n    # Remember, the packet format states that we first receive a length, then packet id, then data\n    _response_len = await conn.read_varint()\n    _response = await conn.read(_response_len)  # will give us a bytearray\n\n    # Amazing, we\'ve just received data from the server! But it\'s just bytes, let\'s turn it into\n    # a Buffer object, which includes helpful methods that allow us to read from it\n    response = Buffer(_response)\n    packet_id = response.read_varint()  # Remember, 2nd field is the packet ID\n\n    # Let\'s see it then, what packet did we get?\n    print(packet_id)  # 0\n\n    # Interesting, this packet has an ID of 0, but wasn\'t that the status request packet? We wanted\n    # a response tho. Well, actually, if we take a look at the status response packet at the wiki,\n    # it really has an ID of 0:\n    # https://wiki.vg/Protocol#Status_Response\n    # Aha, so not only are packet ids unique between game states, they\'re also unique between the\n    # direction a server bound packet (sent by client, with server as the destination) can have an\n    # id of 0, while a client bound packet (sent by server, with client as the destination) can\n    # have the same id, and mean something else.\n\n    # Alright then, so we know what we got is a status response packet, let\'s read the wiki a bit\n    # further and see what data it actually contains, and see how we can get it out. Hmmm, it\n    # contains a UTF-8 encoded string that represents JSON data, ok, so let\'s get that string, it\'s\n    # still in our buffer.\n    received_string = response.read_utf()\n\n    # Now, let\'s just use the json module, convert the string into some json object (in this case,\n    # a dict)\n    data = json.loads(received_string)\n    return data\n\nasync def main():\n    # That\'s it, all that\'s left is actually calling our functions now\n\n    ip = "mc.hypixel.net"\n    port = 25565\n\n    async with (await TCPAsyncConnection.make_client((ip, port), 2)) as connection:\n        await handshake(connection, ip, port)\n        data = await status(connection, ip, port)\n\n    # Wohoo, we got the status data! Let\'s see it\n    print(data["players"]["max"])  # This is the server\'s max player amount (slots)\n    print(data["players"]["online"])  # This is how many people are currently online\n    print(data["description"])  # And here\'s the motd\n\n    # There\'s a bunch of other things in this data, try it out, see what you can find!\n\ndef start():\n    # Just some boilerplate code that can run our asynchronous main function\n    asyncio.run(main())\n```\n\n### Using packet classes for communication\n\n#### Obtaining the packet map\n\nThe first thing you\'ll need to understand about packet classes in mcproto is that they\'re versioned depending on the\nprotocol version you\'re using. As we\'ve already seen with minecraft packets, they\'re following a certain format, and\nfor given packet direction and game state, the packet numbers are unique.\n\nThis is how we can detect what packet is being received, but because of the different versions that the library can\nsupport, we will need to use a packet map, that will contain all of the mappings for given protocol version, from\nwhich, knowing the state and direction, we can get a dictionary with packet IDs as keys, and the individual packet\nclasses as values.\n\nThis dictionary is crucial to receiving packets, as it\'s the only thing that tells us which packet class should be\nused, once we receive a packet and learn about the packet id. Otherwise we wouldn\'t know what to do with the data we\nobtained.\n\nThe first game state we\'ll be in, before doing anything will always be the handshaking state, so let\'s see how we could\ngenerate this dictionary for this state, for all of the receiving (client bound) packets.\n\n```python\nfrom mcproto.packets import PACKET_MAP\nfrom mcproto.packets.abc import PacketDirection, GameState\n\nhandshaking_packet_map = PACKET_MAP.make_id_map(\n    protocol_version=757,\n    direction=PacketDirection.CLIENTBOUND,\n    game_state=GameState.HANDSHAKING\n)\n\nprint(handshaking_packet_map)  # {}\n```\n\nNotice that the packet map is actually empty, and this is simply because there (currently) aren\'t any client bound\npackets a server can send out for the handshaking game state. Let\'s see the status gamestate instead:\n\n```python\nstatus_packet_map = PACKET_MAP.make_id_map(\n    protocol_version=757,\n    direction=PacketDirection.CLIENTBOUND,\n    game_state=GameState.STATUS,\n)\n\nprint(status_packet_map)  # Will print:\n# {1: mcproto.packets.v757.status.ping.PingPong, 0: mcproto.packets.v757.status.status.StatusResponse}\n```\n\nCool! These are all of the packets, with their IDs that the server can send in STATUS game state.\n\n#### Creating our own packets\n\n\nNow, we could create a similar packet map for sending out the packets, and just use it to construct our packets,\nhowever this is not the recommended approach, as it\'s kind of hard to remember all of the packet IDs, and (at least\ncurrently) it means not having any typing information about what packet class will we get. For that reason, it\'s\nrecommended to import packets that we want to send out manually, like so:\n\n```python\nfrom mcproto.packets.v757.handshaking.handshake import Handshake, NextState\n\nmy_handshake = Handshake(\n    # Once again, we use an old protocol version so that even older servers will work\n    protocol_version=47,\n    server_address="mc.hypixel.net",\n    server_port=25565,\n    next_state=NextState.STATUS,\n)\n```\n\nThat\'s it! We\'ve now constructed a full handshake packet with all of the data it should contain. You might remember\nfrom the example above, that we originally had to look at the protocol specification, find the handshake packet and\nconstruct it\'s data as a Buffer with all of these variables.\n\nWith these packet classes, you can simply follow your editor\'s function hints to see what this packet requires, pass it\nin and the data will be constructed for you from these attributes, once we\'ll be to sending it.\n\nFor completion, let\'s also construct the status request packet that we were sending to instruct the server to send us\nback the status response packet.\n\n```python\nfrom mcproto.packets.v757.status.status import StatusRequest\n\nmy_status_request = StatusRequest()\n```\n\nThis one was even easier, as the status request packet alone doesn\'t contain any special data, it\'s just a request to\nthe server to send us some data back.\n\n#### Sending packets\n\nTo actually send out a packet to the server, we\'ll need to create a connection, and use the custom functions\nresponsible for sending packets out. Let\'s see it:\n\n```python\nfrom mcproto.packets import async_write_packet\nfrom mcproto.connection import TCPAsyncConnection\n\nasync def main():\n    ip = "mc.hypixel.net"\n    port = 25565\n\n    async with (await TCPAsyncConnection.make_client((ip, port), 2)) as connection:\n        await async_write_packet(connection, my_handshake)  \n        # my_handshake is a packet we\'ve created in the example before\n```\n\nHow quick was that? Now compare this to the manual version.\n\n#### Receiving packets\n\nAlright, we might now know how to send a packet, but how do we receive one? Let\'s see:\n\n```python\nfrom mcproto.packets import PACKET_MAP\n\n# Let\'s prepare the packet map we\'ll be using, say we\'re already in the STATUS game state now\nSTATUS_PACKET_MAP = PACKET_MAP.make_id_map(\n    protocol_version=757,\n    direction=PacketDirection.CLIENTBOUND,\n    game_state=GameState.STATUS\n)\n\n# Let\'s say we already have a connection at this moment, after all, how else would\n# we\'ve gotten into the STATUS game state.\n# Also, let\'s do something different, let\'s say we have a synchronous connection\nfrom mcproto.connection import TCPSyncConnection\nconn: TCPSyncConnection\n\n# With a synchronous connection, comes synchronous reader, so instead of using async_read_packet,\n# we\'ll use sync_read_packet here\nfrom mcproto.packets import sync_read_packet\n\npacket = sync_read_packet(conn, STATUS_PACKET_MAP)\n\n# Cool! We\'ve got back a packet, but what packet is it? Let\'s import the packet classes it could\n# be and check against them\nfrom mcproto.packets.v757.status.status import StatusResponse\nfrom mcproto.packets.v757.status.ping import PingPong\n\nif isinstance(packet, StatusResponse):\n    ...\nelif isinstance(packet, PingPong):\n    ...\nelse:\n    raise Exception("Impossible, there aren\'t other client bound packets in the STATUS game state")\n```\n\n#### Requesting status\n\nNow let\'s actually do something meaningful, and replicate the entire example from the manual version using packets,\nlet\'s see just how much simpler it will be:\n\n```python\nfrom mcproto.connection import TCPAsyncConnection\nfrom mcproto.packets import async_write_packet, async_read_packet, PACKET_MAP\nfrom mcproto.packets.abc import PacketDirection, GameState\nfrom mcproto.packets.v757.handshaking.handshake import Handshake, NextState\nfrom mcproto.packets.v757.status.status import StatusRequest, StatusResponse\n\nSTATUS_PACKET_MAP = PACKET_MAP.make_id_map(\n    protocol_version=757,\n    direction=PacketDirection.CLIENTBOUND,\n    game_state=GameState.STATUS\n)\n\n\nasync def get_status(ip: str, port: int) -> dict:\n    handshake_packet = Handshake(\n        protocol_version=47,\n        server_address=ip,\n        server_port=port,\n        next_state=NextState.STATUS,\n    )\n    status_req_packet = StatusRequest()\n\n    async with (await TCPAsyncConnection.make_client((ip, port), 2)) as connection:\n        # We start out at HANDSHAKING game state\n        await async_write_packet(connection, handshake_packet)  \n        # After sending the handshake, we told the server to now move us into the STATUS game state\n        await async_write_packet(connection, status_req_packet)  \n        # Since we\'re still in STATUS game state, we use the status packet map when reading\n        packet = await async_read_packet(connection, STATUS_PACKET_MAP)\n\n    # Now that we\'ve got back the packet, we no longer need the connection, we won\'t be sending\n    # anything else. Let\'s just make sure it really is the packet we expected\n    if not isinstance(packet, StatusResponse):\n        raise ValueError(f"We\'ve got an unexpected packet back: {packet!r}")\n\n    # Now that we know we\'re dealing with a status response, let\'s get out it\'s data, and return in\n    return packet.data\n```\n\nWell, that wasn\'t so hard, was it?\n',
    'author': 'ItsDrike',
    'author_email': 'itsdrike@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/py-mine/mcproto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import ClassVar, TYPE_CHECKING

from mcproto.buffer import Buffer
from mcproto.packets.abc import ClientBoundPacket, GameState, ServerBoundPacket
from mcproto.protocol.base_io import StructFormat

if TYPE_CHECKING:
    from typing_extensions import Self


__all__ = ["PingPong"]


class PingPong(ClientBoundPacket, ServerBoundPacket):
    """Ping request/Pong response (Server <-> Client).

    :param int payload: Random number to test out the connection (ideally a long one).
    """

    __slots__ = ("payload",)

    PACKET_ID: ClassVar[int] = 0x01
    GAME_STATE: ClassVar[GameState] = GameState.STATUS

    def __init__(self, payload: int):
        self.payload = payload

    def serialize(self) -> Buffer:
        buf = Buffer()
        buf.write_value(StructFormat.LONGLONG, self.payload)
        return buf

    @classmethod
    def deserialize(cls, buf: Buffer, /) -> Self:
        payload = buf.read_value(StructFormat.LONGLONG)
        return cls(payload)



from __future__ import annotations

import json
from typing import Any, ClassVar, TYPE_CHECKING

from mcproto.buffer import Buffer
from mcproto.packets.abc import ClientBoundPacket, GameState, ServerBoundPacket

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ["StatusRequest", "StatusResponse"]


class StatusRequest(ServerBoundPacket):
    """Request from the client to get information on the server. (Client -> Server)"""

    __slots__ = ()

    PACKET_ID: ClassVar[int] = 0x00
    GAME_STATE: ClassVar[GameState] = GameState.STATUS

    def serialize(self) -> Buffer:
        return Buffer()

    @classmethod
    def deserialize(cls, buf: Buffer, /) -> Self:
        return cls()


class StatusResponse(ClientBoundPacket):
    """Response from the server to requesting client with status data information. (Server -> Client)

    :param dict[str, Any] data: JSON response data sent back to the client.
    """

    __slots__ = ("data",)

    PACKET_ID: ClassVar[int] = 0x00
    GAME_STATE: ClassVar[GameState] = GameState.STATUS

    def __init__(self, data: dict[str, Any]):
        self.data = data

    def serialize(self) -> Buffer:
        buf = Buffer()
        s = json.dumps(self.data)
        buf.write_utf(s)
        return buf

    @classmethod
    def deserialize(cls, buf: Buffer, /) -> Self:
        s = buf.read_utf()
        data_ = json.loads(s)
        return cls(data_)



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, ClassVar, Literal, TYPE_CHECKING, overload

from mcproto.packets.abc import ClientBoundPacket, GameState, Packet, PacketDirection, ServerBoundPacket
from mcproto.utils.version_map import VersionMap, WalkableModuleData

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

__all__ = ["PacketMap"]


class PacketMap(VersionMap["tuple[PacketDirection, GameState, int]", "type[Packet]"]):
    SUPPORTED_VERSIONS: ClassVar[set[int]] = {757}
    COMPATIBLE_FALLBACK_VERSIONS: ClassVar[set[int]] = set()
    _SEARCH_DIR_QUALNAME: ClassVar[str] = "mcproto.packets"

    __slots__ = tuple()

    @overload
    def make_id_map(
        self,
        protocol_version: int,
        direction: Literal[PacketDirection.CLIENTBOUND],
        game_state: GameState,
    ) -> dict[int, type[ClientBoundPacket]]:
        ...

    @overload
    def make_id_map(
        self,
        protocol_version: int,
        direction: Literal[PacketDirection.SERVERBOUND],
        game_state: GameState,
    ) -> dict[int, type[ClientBoundPacket]]:
        ...

    def make_id_map(
        self,
        protocol_version: int,
        direction: PacketDirection,
        game_state: GameState,
    ) -> Mapping[int, type[Packet]]:
        """Construct a dictionary mapping (packet ID -> packet class) for values matching given attributes."""
        res = {}
        for (k_direction, k_game_state, k_packet_id), v in self.make_version_map(protocol_version).items():
            if k_direction is direction and k_game_state is game_state:
                res[k_packet_id] = v
        return res

    def _check_obj(
        self,
        obj: Any,  # noqa: ANN401
        module_data: WalkableModuleData,
        protocol_version: int,
    ) -> TypeGuard[type[Packet]]:
        """Determine whether a member object should be considered as a valid component for given protocol version.

        This method will be called for each potential member object (found when walking over all members of
        __all__ from all submodules of the package for given protocol version).

        This function shouldn't include any checks on whether an object is already registered in the version map
        (key collisions), these are handled during the collection in load_version, all this function is responsible
        for is checking whether this object is a valid component, components with conflicting keys are still
        considered valid here, as they're handled elsewhere.

        Although if there is some additional data that needs to be unique for a component to be valid, which wouldn't
        be caught as a key collision, this function can raise a ValueError.
        """
        return issubclass(obj, Packet)

    @classmethod
    def _make_obtain_key(
        cls,
        obj: type[Packet],
        module_data: WalkableModuleData,
        protocol_version: int,
    ) -> tuple[PacketDirection, GameState, int]:
        """Construct a unique obtain key for given object under given protocol version.

        Note: While the protocol version might be beneficial to know when constructing
        the obtain key, it shouldn't be used directly as a part of the key, as the items
        will already be split by their protocol versions, and this version will be known
        at obtaining time.
        """
        if issubclass(obj, ClientBoundPacket):
            direction = PacketDirection.CLIENTBOUND
        elif issubclass(obj, ServerBoundPacket):
            direction = PacketDirection.SERVERBOUND
        else:
            raise ValueError("Invalid packet class: Neither server-bound not client-bound.")

        return direction, obj.GAME_STATE, obj.PACKET_ID
